<?php

namespace Drupal\cron_queue_invoker;

use Drupal\Component\Datetime\DateTimePlus;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\QueueWorkerManagerInterface;
use Drupal\Core\State\StateInterface;

/**
 * Manage populating queue with jobs automatically on cron.
 *
 * @todo: Cache the applicable queue workers so we don't have to scan them all.
 */
class CronQueueManager {

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * The request time.
   *
   * @var \Drupal\Component\Datetime\DateTimePlus
   */
  protected $requestTime;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The queue worker manager.
   *
   * @var \Drupal\Core\Queue\QueueWorkerManagerInterface
   */
  protected $queueWorkerManager;

  /**
   * The queue factory.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * CronQueueManager constructor.
   *
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\Queue\QueueWorkerManagerInterface $queueWorkerManager
   *   The queue worker manager.
   * @param \Drupal\Core\Queue\QueueFactory $queueFactory
   *   The queue factory.
   */
  public function __construct(TimeInterface $time, StateInterface $state, QueueWorkerManagerInterface $queueWorkerManager, QueueFactory $queueFactory) {
    $this->time = $time;
    $this->requestTime = DateTimePlus::createFromTimestamp($this->time->getCurrentTime());
    $this->state = $state;
    $this->queueWorkerManager = $queueWorkerManager;
    $this->queueFactory = $queueFactory;
  }

  /**
   * Queue items on according to their schedule.
   *
   * @param array|null $plugin_ids
   *   Optionally specific queue worker plugin IDs we want to trigger on
   *   schedule.
   */
  public function onSchedule(?array $plugin_ids = NULL): void {
    $definitions = $this->queueWorkerManager->getDefinitions();

    // Filter by the provided plugin IDs, if any.
    if ($plugin_ids) {
      $definitions = array_filter($definitions, function (array $definition) use ($plugin_ids) {
        return in_array($definition['id'], $plugin_ids);
      });
    }

    // Filter by those that should be scheduled.
    $definitions = array_filter($definitions, [$this, 'shouldWorkerBeStarted']);

    // Create the queue items.
    $this->startQueueWorkers($definitions);
  }

  /**
   * Start the given queue workers.
   *
   * @param array[]|string[] $plugins
   *   Optionally specific queue worker plugin IDs or plugin definitions we want
   *   to trigger on schedule.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function startQueueWorkers(?array $plugins = NULL): void {
    if (!$plugins) {
      $plugins = $this->queueWorkerManager->getDefinitions();
    }

    foreach ($plugins as $plugin) {
      // Ensure we have the plugin definition.
      if (!is_array($plugin)) {
        $plugin = $this->queueWorkerManager->getDefinition($plugin);
      }

      // Check the plugin should be invoked on cron.
      if (!$plugin['cron_invoke']) {
        continue;
      }

      // Add a queue item with the last run date.
      $this->queueFactory
        ->get($plugin['id'])
        ->createItem([$this->getLastRunTime($plugin), $this->requestTime]);

      // Set the last run time.
      $this->setLastRunTime($plugin);
    }
  }

  /**
   * Check whether a queue worker should be started according to it's schedule.
   *
   * @param array $definition
   *   The plugin definition.
   *
   * @return bool
   *   Whether an item should be queued.
   *
   * @throws \Exception
   *   If the interval defined in the cron queue is invalid.
   *
   * @todo: Support intervals in the format supported by \DateInterval().
   */
  protected function shouldWorkerBeStarted(array $definition): bool {
    // Check the plugin should be invoked on cron.
    if (!$definition['cron_invoke']) {
      return FALSE;
    }

    // If there was no last run, we invoke immediately.
    $last_run = $this->getLastRunTime($definition);
    if (!$last_run) {
      return TRUE;
    }

    // Consistently format our times based on the desired interval.
    $format = 'Y-m-d H:i:s';
    $split_pos = strpos($format, $definition['cron_invoke']['interval']);
    if ($split_pos === FALSE) {
      throw new \Exception('Invalid run interval.');
    }
    $run_format = trim(substr($format, 0, $split_pos + 1), '- ');

    // If we have a last run, check it was before today.
    if ($last_run->format($run_format) >= $this->requestTime->format($run_format)) {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Get the last run time for a queue worker.
   *
   * @param array $definition
   *   The plugin definition.
   *
   * @return \Drupal\Component\Datetime\DateTimePlus|null
   *   The last run time, or NULL for never.
   */
  protected function getLastRunTime(array $definition): ?DateTimePlus {
    $state = $definition['cron_invoke']['state'];
    $last_run_timestamp = $this->state->get($state);
    return isset($last_run_timestamp) ? DateTimePlus::createFromTimestamp($last_run_timestamp) : NULL;
  }

  /**
   * Set the last run time for a queue worker.
   *
   * @param array $definition
   *   The plugin definition.
   */
  protected function setLastRunTime(array $definition): void {
    $state = $definition['cron_invoke']['state'];
    $this->state->set($state, $this->requestTime->getTimestamp());
  }
}
